
/* "Species" - a CoreWars evolver.  Copyright (C) 2003 'Varfar' 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 1, or (at your option) any later
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA. 
 */
 
#include "genus.hpp"
using namespace std;

/***** CGenus class implementation *********************/

CGenus::CGenus(CKingdom *kingdom,const string &name): CHookable(), _kingdom(kingdom), _name(name) {
	// set some defaults to inherit from the parent
	_fitness = _kingdom->fitness();
	_length = _kingdom->length();
	_freq = _kingdom->freq();
	_operands = _kingdom->operands();
	_reproduction = _kingdom->reproduction();
	_species = 0;
	_num_species = 0;
}

CGenus::~CGenus() {
	// delete all species
	clear();
	// delete anything not inherited
	if(kingdom()->fitness() != _fitness)
		delete _fitness;
	if(kingdom()->length() != _length)
		delete _length;
	if(kingdom()->freq() != _freq)
		delete _freq;
	if(kingdom()->operands() != _operands)
		delete _operands;
	if(kingdom()->reproduction() != _reproduction)
		delete _reproduction;
}

void CGenus::clear() { /* removes all species from this genus */
	int i;
	if(0 == _species) // not allocated?
		return;
	for(i=0; i<_num_species; i++) {
		delete _species[i];
	}
	_num_species = 0;
	delete[] _species;
	_species = 0;
}

int CGenus::num_evolving_species() const {
	int i, count=0;
	for(i=0; i<_num_species; ++i)
		if(_species[i]->is_evolving())
			count++;
	return count;
}

int CGenus::num_evolved_warriors() const {
	int i, count=0;
	for(i=0; i<_num_species; ++i)
		if(_species[i]->is_evolving())
			count += _species[i]->num_warriors();
	return count;
}

int CGenus::num_warriors() const { /* returns a trivia number of warriors 'live' in this genus at this time */
	int i, count=0;
	for(i=0; i<_num_species; ++i)
		count += _species[i]->num_warriors();
	return count;
}

void CGenus::read_ini(INIFile &ini) {
	KeyValuePair *kvp;
	int i;
	string num;
	clear();
	// load some defaults; these might be overloaded
	_fitness = _fitness->read_override_ini(ini);
	_length = _length->read_override_ini(ini);
	_freq = CInstGenerator::read_override_ini(_freq,ini);
	_operands = COperand::read_override_ini(_operands,ini);
	_reproduction = _reproduction->read_override_ini(ini);
	// how many species do we have?
	kvp = ini.get("num_species"); if(0 == kvp) PANIC(MISC,"num_species expected",NULL); _num_species = kvp->getValueAsInt();
	_species = new CSpecies*[_num_species];
	// load each species name
	for(i=0; i<_num_species; ++i) {
		// prepare the number of this species as a string to search for
		num = "";
		string_append(num,i);
		// and get it
		kvp = ini.get(num); if (0 == kvp) PANIC(MISC,"species expected",NULL);
		// and create it
		if(verbose) { // debug output
			cout << "\tadding " << kvp->getValueAsChar() << endl;
		}
		_species[i] = new CSpecies(this,kvp->getValueAsChar(),false);
	}
	// actually load each species
	for(i=0; i<_num_species; ++i) {
		// go to the correct section
		if(verbose) { // debug output
			cout << "\tloading " << _species[i]->name() << endl;
		}
		if(!ini.seek(_species[i]->name())) PANIC(MISC,_species[i]->name().c_str(),"species section not found");
		// and load it
		_species[i]->read_ini(ini);
	}	
}

void CGenus::write_ini(ostream &os) {
	int i;
	// dump misc stuff
	_fitness->write_override_ini(os,_kingdom->fitness());
	_length->write_override_ini(os,_kingdom->length());
	_freq->write_override_ini(os,_kingdom->freq());
	_operands->write_override_ini(os,_kingdom->operands());
	_reproduction->write_override_ini(os,_kingdom->reproduction());
	// list species
	os << "num_species=" << _num_species << endl;
	for(i=0; i<_num_species; ++i) {
		os	<< i << '=' << _species[i]->name() << endl;
	}
	// dump the bodies of all the species declared
	for(i=0; i<_num_species; ++i) {
		os	<< endl
			<< '[' << _species[i]->name() << ']' << endl;
		_species[i]->write_ini(os);
	}
}

void CGenus::save() {
	int i;
	for(i=0; i<_num_species; ++i) {
		_species[i]->save();
	}
}

void CGenus::ready() { /* this will blank all fitness metrics preceeding the fighting of a generation */
	int i;
	for(i=0; i<_num_species; ++i) {
		_species[i]->ready();
	}
}

int CGenus::fight(CGenus &enemy) { /* this will actually do all the fights for a generation; returns number of rounds fought */
	int i, j, count = 0;
	for(i=0; i<_num_species; ++i) {
		if(_species[i]->is_bench()) // only if it is an evolving species; no point gathering WINNING points for benchmarks benchmarks
			continue;
		for(j=0; j<enemy._num_species; ++j) {
			if(verbose) { // debug output
				cout << "\tfight: " << _species[i]->name() << " vs. " << enemy._species[j]->name() << endl;
			}
			count += _species[i]->fight(*enemy._species[j]);
		}
	}
	return count;
}

CWarrior *CGenus::selection() { /* this will do selection on all the warriors in this genus; returns pointer to the best */
	int i;
	CWarrior *best = 0, *candidate;
	for(i=0; i<_num_species; ++i) {
		if(verbose) {
			cout << "breeding " << _species[i]->uid() << endl;
		}
		candidate = _species[i]->selection();
		if(0 == best) // not set?
			best = candidate;
		else if(0 != candidate) // this species did return a best (so is evolving)
			if(candidate->score() > best->score()) // beaten
				best = candidate;
	}
	return best;
}

CWarrior *CGenus::warrior(const CWarrior::TUid uid) const { /* fetch a particular warrior from this genus; NULL if not present */
	int i;
	CWarrior *ret = 0;
	for(i=0; (i < _num_species) && (0 == ret); ++i) {
		ret = _species[i]->warrior(uid);
	}
	return ret;
}

//*** array index accessors

CSpecies &CGenus::species(unsigned index) const {
	if(index >= _num_species)
		PANIC(MISC,"illegal CGenus species index",NULL);
	return *_species[index];
}
