
#ifndef HOOK_HPP
#define HOOK_HPP

/* "Species" - a CoreWars evolver.  Copyright (C) 2003 'Varfar' 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 1, or (at your option) any later
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA. 
 */
 
// forwards
class CSpecies;
class CGenus;
class CHookable;

/** class CHook
 * This thin class provides a way to attach event listeners or information to objects in the Species hierarchy.
 * Each genus and species can have a single hook attached.
 * Important: The genus or species that has the hook attached 'owns' it.  You should not delete hooks elsewhere.
 * When you assign a hook (or *0) to an object, the old hook will be deleted if set. 
 **/
class CHook {
	friend class CHookable;
	public:
		virtual ~CHook() {}
		// actual event callbacks; override to listen in
	protected:
		virtual void on_fight(const CSpecies &/*self*/,const CSpecies &/*other*/) {}
		virtual void on_fight(const CGenus &/*self*/,const CGenus &/*other*/) {}
		virtual void on_reproduction(const CSpecies &/*self*/) {}
		virtual void on_reproduction(const CGenus &/*self*/) {}
};

/** class CHookable
 * to make an object 'hookable', simply derive from this class.  In your code, call the appropriate notifiers
 **/
class CHookable {
	public:
		CHook *hook() const { return _hook; }
		bool has_hook() const { return 0 != _hook; }
		void hook(CHook *newhook) { delete _hook; _hook = newhook; }
	protected:
		CHookable() { _hook = 0; }
		~CHookable() { delete _hook; }
		void notify_hook_fight(const CSpecies &self,const CSpecies &other) {
			if(0 != _hook)
				_hook->on_fight(self,other);
		}
		void notify_hook_fight(const CGenus &self,const CGenus &other) {
			if(0 != _hook)
				_hook->on_fight(self,other);
		}
		void notify_reproduction(const CSpecies &self) {
			if(0 != _hook)
				_hook->on_reproduction(self);
		}
		void notify_reproduction(const CGenus &self) {
			if(0 != _hook)
				_hook->on_reproduction(self);
		}
	private:
		CHook *_hook;
};

#endif // ifndef HOOK_HPP


