
/* "Species" - a CoreWars evolver.  Copyright (C) 2003 'Varfar' 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 1, or (at your option) any later
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include "ini.hpp"

#include <string.h>

using namespace std;

namespace { // anonymous 

	const char *KVP_ERRORS[KeyValuePair::KVP_ERROR_LAST] = {
    	"VALUE_NOT_INT","VALUE_NOT_BOOL","VALUE_NOT_FLOAT",
    	"VALUE_NOT_DOUBLE","KEY_NOT_VALID","KEY_HAS_NO_CATEGORY"
	};
	
	bool str_equals(const string str1,const char *s2) {
		const char *s1 = str1.data(); 
		unsigned int s11, s12, s21, s22;
		for(s11=0; s11<strlen(s1); s11++) if(s1[s11] > ' ') break;
		for(s12=strlen(s1)-1; s12 > s11; s11--) if(s1[s12] > ' ') break;
		for(s21=0; s21<strlen(s2); s21++) if(s2[s21] > ' ') break;
		for(s22=strlen(s2)-1; s22 > s21; s22--) if(s2[s22] > ' ') break;
		// easy length check
		if((s12-s11) != (s22-s21)) return false;
		// ok, have to compare each character
		return (strncmp(&s1[s11],&s2[s21],s12-s11) == 0);
	}

} // end of anon namespace

KeyValuePair::KeyValuePair() {
   linenum = 0;
   valid = false;
}

KeyValuePair::~KeyValuePair() {}

const char *KeyValuePair::ERROR_MSG(const KVP_ERROR err) {
	if(0 <= err < KVP_ERROR_LAST)
		return KVP_ERRORS[err];
	return 0;
}

// checking key

bool KeyValuePair::is(const char *key) const {
   if(!valid) return false; // obviously
   return (strcasecmp(this->key,key) == 0);
}

// getting key

void KeyValuePair::getKeyAsString(string &key) {
   if(!valid) throw KEY_NOT_VALID;
   key = string(this->key); // reconstruct
}

const char *KeyValuePair::getKeyAsChar() {
   if(!valid) throw KEY_NOT_VALID;
   return key;
}

// checking category

bool KeyValuePair::isCategory(const char *category) {
   if(!valid) return false; // obviously
   if(!hascategory) return false; // obviously
   return (strcasecmp(this->category,category) == 0);
}

bool KeyValuePair::hasSubCategory() { // more categories?  category.category...key
   for(char *ch=key; *ch != '\0'; ch++)
      if(*ch == '.') return true;
   return false;
}

bool KeyValuePair::getSubCategory() { // will divide the current key into category.key
   hascategory = false;
   char *ch;
   for(ch = key; *ch != '\0'; ch++)
      if(*ch == '.') { // find dot delimiter
         hascategory = true;
         break;
      }
   if(hascategory) {
      category = key;
      *ch = '\0';
      key = ++ch;
      //cout << "category = \"" << category << "\"\tkey=\"" << key << '\"' << endl;
      if(*key == '\0') { // empty key now?  was trailing . ?
         // repair damage!
         hascategory = false;
         *--ch = '.';
         key = category;
      }
   } else
      category = NULL;
   return hascategory;
}

// getting category

void KeyValuePair::getCategoryAsString(string &category) {
   if(!valid) throw KEY_NOT_VALID;
   if(!hascategory) throw KEY_HAS_NO_CATEGORY;
   category = string(this->category); // reconstruct
}

const char *KeyValuePair::getCategoryAsChar() const {
   if(!valid) throw KEY_NOT_VALID;
   if(!hascategory) throw KEY_HAS_NO_CATEGORY;
   return category;
}

// checking value

bool KeyValuePair::isValue(const char *val) const {
	if(!valid) return false; // obviously
	return (strcasecmp(this->value,val) == 0);
}

// getting value

void KeyValuePair::getValueAsString(string &value) {
   if(!valid) throw KEY_NOT_VALID;
   value = string(this->value); // reconstruct
}

const char *KeyValuePair::getValueAsChar() {
   if(!valid) throw KEY_NOT_VALID;
   return value;
}

int KeyValuePair::getValueAsInt(int def) {
   // Will's patent parser detects errors better than atoi
   if(!valid) throw KEY_NOT_VALID;
   char *ch = value;
   if('\0' == *ch) return def;
   bool neg = (*ch=='-'); // remember if negative
   if(*ch=='+' || *ch=='-') // step past sign?
      ch++;
   int result = 0;
   while(*ch>='0' && *ch<='9') {
      int digit = (*ch)-'0'; // 0..9
      result *= 10;
      result += digit;
      ch++;
   }
   if(*ch!='\0') throw VALUE_NOT_INT;
   if(neg)
      result = -result;
   return result;
}

float KeyValuePair::getValueAsFloat(float def,bool percent_allowed) {
   // Will's patent parser detects errors better than atof
   if(!valid) throw KEY_NOT_VALID;
   char *ch = value;
   if('\0' == *ch) return def;
   bool neg = (*ch=='-'); // remember if negative
   if(*ch=='+' || *ch=='-') // step past sign?
      ch++;
   float result = 0.0f;
   int digit;
   while(*ch>='0' && *ch<='9') {
      digit = (*ch)-'0'; // 0..9
      result *= 10;
      result += digit;
      ch++;
   }
   if(*ch=='.' || *ch==',') { // floating point?  Note European-friendly :)
      ch++; // step past it
      float weight = 1.0f;
      while(*ch>='0' && *ch<='9') {
         digit = (*ch)-'0'; // 0..9
         weight *= 10.0f;
         result += digit/weight;
         ch++;
      }
   }
   if(*ch=='%') {
      if(percent_allowed)
         result = result / 100.0f;
      else
      	throw VALUE_NOT_FLOAT;
      ch++; // step over it
   }
   if(*ch!='\0')
   	throw VALUE_NOT_FLOAT;
   if(neg)
      result = -result;
   return result;
}

double KeyValuePair::getValueAsDouble(double def,bool percent_allowed) {
   // Will's patent parser detects errors better than atof
   if(!valid) throw KEY_NOT_VALID;
   char *ch = value;
   if('\0' == *ch) return def;
   bool neg = (*ch=='-'); // remember if negative
   if(*ch=='+' || *ch=='-') // step past sign?
      ch++;
   double result = 0.0;
   int digit;
   while(*ch>='0' && *ch<='9') {
      digit = (*ch)-'0'; // 0..9
      result *= 10;
      result += digit;
      ch++;
   }
   if(*ch=='.' || *ch==',') { // floating point?  Note European-friendly :)
      ch++; // step past it
      float weight = 1.0;
      while(*ch>='0' && *ch<='9') {
         digit = (*ch)-'0'; // 0..9
         weight *= 10.0;
         result += digit/weight;
         ch++;
      }
   }
   if(*ch=='%') {
      if(percent_allowed)
         result = result / 100.0;
      else
      	throw VALUE_NOT_DOUBLE;
      ch++; // step over it
   }
   if(*ch!='\0')
   	throw VALUE_NOT_DOUBLE;
	if(neg)
      result = -result;
   return result;
}

bool KeyValuePair::getValueAsBool(bool def) {
   if(!valid) throw KEY_NOT_VALID;
   if('\0' == *value) return def;
   if(strcasecmp(value,"true")==0 || strcasecmp(value,"on")==0 || strcasecmp(value,"1")==0 || strcasecmp(value,"yes")==0)
      return true;
   if(strcasecmp(value,"false")==0 || strcasecmp(value,"off")==0 || strcasecmp(value,"0")==0 || strcasecmp(value,"no")==0)
      return false;
   throw VALUE_NOT_BOOL;
}

// read line from input stream

// uses very c-style parsing techniques with character pointers
istream &KeyValuePair::fromSerial(istream &is) {
   valid = false;
   while(!is.eof()) {
      // get the line
      is.getline(line,MAX_KVP_LINE_LEN);
      linenum++;
      //cout << "line #" << linenum << " is \"" << line << '\"' << endl;
      // trim leading whitespace
      key = line;
      while(*key==' ' || *key=='\t')
         key++;
      // is blank line?
      if(*key=='\0' || *key=='\r' || *key==';' || *key=='#') // empty or full-line comment?
         continue; // try again
      // is another section?
      if(*key=='[')
      	break; // abort parse
      // truncate at any trailing comment
      char *end = key;
      while(*end!=';' && *end!='#' && *end!='\0' && *end!='\r')
         end++;
      *end = '\0'; // do truncate
      // work out split between key and value
      value = key;
      while(*value!='=' && *value!='\0')
         value++;
      if(*value=='\0') // no separator?
         continue;
      // trim key
      char *ch = value;
      while(ch>key && (*ch==' ' || *ch=='\t' || *ch=='='))
         ch--;
      ch++;
      *ch = '\0'; // truncate; key done
      if(*key=='\0') // check again
         continue;
      // check key for illegal characters
      bool illegal = false;
      for(ch=key; *ch != '\0'; ch++)
         if(!((*ch>='A' && *ch<='Z') || (*ch>='a' && *ch<='z') ||
            (*ch>='0' && *ch<='9') || *ch=='.' || *ch=='-' || *ch=='_')) {
            cerr << "ini: illegal char (" << int(*ch) << ") in key on line " << linenum << endl;
            illegal = true;
            break;
         }
      if(illegal) continue;
      // trim whitespace from leading value
      value++; // move past =
      while(*value==' ' || *value=='\t')
         value++;
      if(*value=='\0') { // value is empty?
         cerr << "ini: value is empty on line " << linenum << " (" << key << ')' << endl;
         continue;
      }
      // trim whitespace from trailing value
      ch = end;
      ch--;
      while(ch>value && (*ch==' ' || *ch=='\t'))
         ch--;
      ch++;
      *ch = '\0'; // truncate here too
      // check value for illegal characters
      illegal = false;
      for(ch=value; *ch != '\0'; ch++)
         if(*ch<32 || *ch>126) { // 7-bit printable ASCII
            cerr << "ini: illegal char (" << int(*ch) << ") in value on line " << linenum << endl;
            illegal = true;
            break;
         }
      if(illegal) continue;
      // category in key?
      getSubCategory();
      // done
      valid = true;
      break;
   }
   return is;
}

/******* INIFile class implementation ***********************/

INIFile::INIFile(istream &is): _is(&is), _section_ofs(NO_SECTION), _kvp() {
	// find out if we have an empty stream
	_is->seekg(0,ios::end); // go to the bottom of the stream
	_bad_stream = (0 >= _is->tellg());
}

bool INIFile::seek(const string &section) { // returns success
	if(_bad_stream)
		return false;
	string wanted = "["; wanted += section; wanted += ']';
	_is->clear(); // clear any flags
	_is->seekg(0,ios::beg); // go to top of stream
	while(!_is->eof()) {
		_is->getline(_section,KeyValuePair::MAX_KVP_LINE_LEN);
		//cout << "x: " << _section << endl;
		if(str_equals(wanted,_section)) { // found?
			_section_ofs = _is->tellg();
			_section_ofs--; // just step back a bit to try to fix obscure bug
			return true;
		}
	}
	_section_ofs = NO_SECTION;
	return false; // not found
}

KeyValuePair *INIFile::get(const string &key) { // returns null if no such key
	if(NO_SECTION == _section_ofs) // bad section?
		return 0;
	_is->clear(); // clear any flags
	_is->seekg(_section_ofs,ios::beg); // seek to beginning of section data
	_kvp = KeyValuePair();
	*_is >> _kvp;
	while(_kvp.isValid()) {
		if(_kvp.is(key)) // match?
			return &_kvp;
		*_is >> _kvp; // next
	}
	return 0; // bad
}

