
/* "Species" - a CoreWars evolver.  Copyright (C) 2003 'Varfar' 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 1, or (at your option) any later
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include "species.hpp"
#include "rand.hpp"

#include "exec_trail.hpp"

#include <time.h>

#include <iostream>
#include <fstream>
#include <iomanip>
using namespace std;

#ifdef SHOW_PROGRESS_TWIDDLE
int progress = 0;
const char *progress_twiddle = "-\\|/";
#endif // ifdef SHOW_PROGRESS_TWIDDLE

/***** CResumeFile class declaration/implementation ********/

class CResumeFile {
	public:
		CResumeFile(CSpecies *species): _species(species) {
			// name of this file?
			_name = "species-";
			_name += species->name();
			_name += "-resume.dat";
			_write_count = 0;
		}
		bool openForRead() {
			CGeneration::NUM gen;
			_write_count = 0;
			// try to open existing file..
			_file.clear();
			_file.open(_name.c_str(),ios::in|ios::out|ios::binary);
			if(!_file.is_open()) return false;
			// check generation
			_file.read((char*)&gen,sizeof(gen));
			if(!_file.good() || (_species->kingdom()->generation() != gen)
				|| _file.eof()) {
				close();
				return false;
			}
			return true;
		}
		void openForWrite() { // write
			CGeneration::NUM gen = _species->kingdom()->generation();
			_write_count = 0;
			// (re)create file
			_file.clear();
			_file.open(_name.c_str(),ios::out|ios::binary);
			// write generation
			_file.write((char*)&gen,sizeof(gen));
			flush();
		}
		void swapFromReadToWrite() {
			_write_count = 0;
			_file.clear();
		}
		bool read(CWarrior::FIGHT_DATA &fight) {
			if(!_file.good())
				return false;
			_file.read((char*)&fight,sizeof(fight));
			return _file.good();
		}
		void write(const CWarrior::FIGHT_DATA &fight) {
			if(_file.good()) {
				_file.write((const char*)&fight,sizeof(fight));
				if(++_write_count >= WRITE_FLUSH) {
					flush();
					_write_count = 0;
				}
			}
		}
		void flush() {
			if(_file.good()) {
				_file.flush();
			}
		}
		void close() {
			_file.close();
		}
	private:
		static const int WRITE_FLUSH = 15; // every n records, flush
		CSpecies *_species;
		string _name;
		fstream _file;
		int _write_count;
};

	
/***** CSpeciesFile class implementation *******************/

CSpeciesFile::CSpeciesFile(const string species,const CGeneration::NUM generation) {
	set_filename(species,generation);
	open_file();
}

CSpeciesFile::CSpeciesFile(const CSpecies &species,const CGeneration::NUM generation) {
	set_filename(species.name(),generation);
	open_file();
}

CSpeciesFile::~CSpeciesFile() {
	fclose(_f);
}

void CSpeciesFile::set_filename(const string name,const CGeneration::NUM generation) {
	// build filename
	_filename = "species-";
	_filename += name;
	_filename += '-';
	string_append(_filename,generation);
	_filename += ".dat";
}

void CSpeciesFile::open_file() {
	// open file
	_f = fopen(_filename.c_str(),"r+b");
	if(0 == _f) { // can't open file, probably doesn't exist?
		_f = fopen(_filename.c_str(),"w+b");
		if(0 == _f) { // couldn't create a file either
			PANIC(CANNOT_OPEN_GENERATION_FILE,_filename.c_str(),NULL);
		}
	}
}

void CSpeciesFile::writeWarrior(const CWarrior &warrior) {
	unsigned int i, j;
	CFitness::NUM num;
	HEADER header;
	// prepare header
	header.options = HEADER::NONE;
	if(0 < warrior.info().length())
		header.options |= HEADER::HAS_INFO;
	if(warrior.is_benchmarked())
		header.options |= HEADER::HAS_BENCH;
	header.uid = warrior.uid();
	header.generation = warrior.generation();
	header.mother = warrior.mother();
	header.father = warrior.father();
	header.start = warrior.start_idx();
	header.num_chromosomes = warrior.species()->num_chromosomes();
	// write header
	if(1 != fwrite(&header,_header_size,1,_f))
		PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write header");
	// write optional parts
	if(HEADER::HAS_INFO == (header.options & HEADER::HAS_INFO)) {
		i = warrior.info().length()+1; // remember space for \0
		if(MAX_INFO_LEN <= i)
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"info length too long");
		if(1 != fwrite(&i,sizeof(i),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write info length");
		if(i != fwrite(warrior.info().c_str(),1,i,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write info");		
	}
	if(HEADER::HAS_BENCH == (header.options & HEADER::HAS_BENCH)) {
		i = warrior.benchmark_count();
		if(1 != fwrite(&i,sizeof(i),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write benchmark_count");
		num = warrior.benchmark();
		if(1 != fwrite(&num,sizeof(num),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write benchmark");
	}
	// write body
	for(i=0; i<header.num_chromosomes; ++i) {
		CChromosome &chromosome = *warrior.chromosome(i);
		j = chromosome.len(); // reuse this!
		if(1 != fwrite(&j,sizeof(j),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write chromosome length");
		for(j=0; j<chromosome.len(); ++j) { 
			if(1 != fwrite(&chromosome.code(j),_insn_size,1,_f))
				PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not write body");
		}
	}
}

void CSpeciesFile::readWarrior(CWarrior &warrior) {
	unsigned int i, j, clen;
	CFitness::NUM num;
	HEADER header;
	// read header
	if(1 != fread(&header,_header_size,1,_f))
		PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read header");
	// prepare warrior
	warrior.set_uid(header.uid);
	warrior.set_generation(header.generation);
	warrior.set_parents(header.mother,header.father);
	warrior.set_start(header.start);
	if(warrior.species()->num_chromosomes() != header.num_chromosomes)
		PANIC(MISC,"bad number of chromosomes in file",NULL);
	// read optional parts
	if(HEADER::HAS_INFO == (header.options & HEADER::HAS_INFO)) {
		if(1 != fread(&i,sizeof(i),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read info length");
		if(MAX_INFO_LEN <= i)
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"info length too long");
		if(i != fread(_info,1,i,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read info");
		warrior.set_info(_info);
	} else {
		warrior.clear_info();
	}
	if(HEADER::HAS_BENCH == (header.options & HEADER::HAS_BENCH)) {
		if(1 != fread(&i,sizeof(i),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read benchmark count");
		if(1 != fread(&num,sizeof(num),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read benchmark sum");
		warrior.set_benchmark(num,i);
	} else
		warrior.set_benchmark(0,0);
	// read body
	for(i=0; i<header.num_chromosomes; ++i) {
		CChromosome &chromosome = *warrior.chromosome(i);
		if(1 != fread(&clen,sizeof(clen),1,_f))
			PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read chromosome length");
		chromosome.set_len(clen);		
		for(j=0; j<clen; j++) {
			if(1 != fread(&chromosome.code(j),_insn_size,1,_f))
				PANIC(GENERATION_FILE_ERROR,_filename.c_str(),"could not read body");
		}
	}
}
	
/***** CSpecies class implementation *******************/

CSpecies::TUid CSpecies::_uid_seq = 0; // init it
const CFitness::NUM CSpecies::DEFAULT_FITNESS_WEIGHTING = 1.0F; // no affect

CSpecies::CSpecies(CGenus *genus,const string &name,const bool evolving): // create an empty species
	CHookable(), _name(name), _evolving(evolving), _genus(genus) {
	_uid = ++_uid_seq; // set unique uid
	/* inherit defaults */
	_fitness_weighting = DEFAULT_FITNESS_WEIGHTING;
	_fitness = _genus->fitness();
	_length = _genus->length();
	_freq = _genus->freq();
	_operands = _genus->operands();
	_reproduction = _genus->reproduction();
	_exec_trail = 0;
	_chromosome = 0;
	_num_chromosomes = 0;
	_start_chromosome = 0;
	_warrior = 0;
	_next = 0;
	_num_warriors = 0;
	_survives = 0;
	_resume = 0;
	_resuming = false;
}

CSpecies::~CSpecies() {
	// delete all species
	clear();
	delete _exec_trail; //TODO: ophaned COpcodeMap
	// delele chromosomes
	for(int i=0; i<_num_chromosomes; i++)
		delete _chromosome[i];
	// delete anything not inherited
	if(genus()->fitness() != _fitness)
		delete _fitness;
	if(genus()->length() != _length)
		delete _length;
	if(genus()->freq() != _freq)
		delete _freq;
	if(genus()->operands() != _operands)
		delete _operands;
	if(genus()->reproduction() != _reproduction)
		delete _reproduction;
	if(0 != _resume)
		delete _resume;
}

void CSpecies::clear() { /* removes all warriors from this species */
	int i;
	if(0 == _warrior) // not allocated?
		return;
	for(i=0; i<_num_warriors; i++) {
		delete _warrior[i];
	}
	_num_warriors = 0;
	delete[] _warrior;
	delete[] _survives;
	delete[] _next; // pointers not 'owned'
	delete[] _chromosome;
	_warrior = 0;
	_next = 0;
	_survives = 0;
}

int CSpecies::find(const CWarrior::TUid uid) const { // -1 if not found in this species
	int i;
	for(i=_num_warriors-1; i>=0; --i) { // seek backwards, since more recent warriors are more likely to be interesting
		if(uid == _warrior[i]->uid())
			return i; 
	}
	return -1; // not found
}

void CSpecies::read_ini(INIFile &ini) {
	int i, tries;
	string num;
	CSpeciesFile *file;
	KeyValuePair *kvp;
	clear();
	// uid
	kvp = ini.get("uid"); if(0 == kvp) { // not specified?
		if(0 != kingdom()->generation()) { // not first generation? this is an error
			PANIC(MISC,name().c_str(),"uid expected");
		}
	} else { // uid specified
		_uid = kvp->getValueAsInt();
	}
	// evolving?
	kvp = ini.get("evolving"); if(0 == kvp) PANIC(MISC,"evolving expected",NULL);
	_evolving = kvp->getValueAsBool();
	// load some defaults; these might be overloaded
	_fitness = _fitness->read_override_ini(ini);
	_length = _length->read_override_ini(ini);
	_freq = CInstGenerator::read_override_ini(_freq,ini);
	_operands = COperand::read_override_ini(_operands,ini);
	_reproduction = _reproduction->read_override_ini(ini);
	// exec trail specified?
	kvp = ini.get("exec_trail");
	if(0 != kvp) {
		_exec_trail = new CExecTrail(new COpcodeMap()); // hmm
		if(!_exec_trail->load_from_file(kvp->getValueAsChar()))
			PANIC(MISC,_exec_trail->filename().c_str(),"could not load exec trail from file");
	}
	// fitness weighting specified?
	kvp = ini.get("fitness_weighting");
	if(0 != kvp) {
		_fitness_weighting = kvp->getValueAsFloat();
	}
	// split into several chromosomes?
	kvp = ini.get("num_chromosomes");
	if(!_evolving) {
		_auto_chromosome = true; // avoids lots of writing/reading of ini entries later etc
		if(0 != kvp)
			PANIC(MISC,"chromosomes specified for a benchmark warrior",NULL);
	} else if(0 != kvp) {
		_auto_chromosome = false;
		_num_chromosomes = kvp->getValueAsInt();
		if((_num_chromosomes <= 0) || (_num_chromosomes >= length()->max()))
			PANIC(MISC,name().c_str(),"invalid number of chromosomes specified");
		_chromosome = new CChromosomeType*[_num_chromosomes];
		for(i=0; i<_num_chromosomes; i++) {
			// prepare the number of this chromosome as a string to search for
			num = "chromosome_";
			string_append(num,i);
			// and get it
			kvp = ini.get(num); if (0 == kvp) PANIC(MISC,name().c_str(),"chromosome expected");
			// and create it
			if(verbose) { // debug output
				cout << "\t\tadding " << kvp->getValueAsChar() << endl;
			}
			_chromosome[i] = new CChromosomeType(this,kvp->getValueAsChar());
		}
		// start specified?
		kvp = ini.get("start_chromosome");
		if(0 != kvp) {
			_start_chromosome = kvp->getValueAsInt();
			if(_start_chromosome > _num_chromosomes)
				PANIC(MISC,_name.c_str(),"start chromosome is greater than the number of chromosomes in the specices");
		}
	} else {
		_auto_chromosome = true;
		_num_chromosomes = 1;
		_chromosome = new CChromosomeType*[1];
		_chromosome[0] = new CChromosomeType(this);
	}
	// how many warriors?
	kvp = ini.get("num_warriors"); if(0 == kvp) PANIC(MISC,"num_warriors expected",NULL);
	_num_warriors = kvp->getValueAsInt();
	_warrior = new CWarrior*[_num_warriors];
	_next = new CWarrior*[_num_warriors];
	_survives = new bool[_num_warriors];
	// load chromosomes
	if(!_auto_chromosome) {
		for(i=0; i<_num_chromosomes; i++) {
			if(verbose) { // debug output
				cout << "\t\tloading " << _chromosome[i]->name() << endl;
			}
			_chromosome[i]->read_ini(ini);
		}
	}
	// load them if benchmark
	if(is_bench()) {
		for(i=0; i<_num_warriors; ++i) {
			// prepare index number
			num = "warrior_";
			string_append(num,i);
			// get it
			kvp = ini.get(num); if(0 == kvp) { // hmm
				PANIC(MISC,num.c_str(),"expected");
			} else { 
				kvp->getValueAsString(num); // reuse string
				if(verbose) { // debug output
					cout << "\t\tbench " << i << '=' << num << endl;
				}
				_warrior[i] = new CWarrior(this,num);
				_warrior[i]->code_check();
			}
		}
	} else { // evolving
		num = "species-";
		num += _name;
		num += "-";
		string_append(num,kingdom()->generation());
		num += ".dat";
		if(file_exists(num) || (0 != kingdom()->generation())) { // not first generation, so load them?
			file = new CSpeciesFile(*this,kingdom()->generation());
			for(i=0; i<_num_warriors; ++i) {
				_warrior[i] = new CWarrior(this,0); // uid will be overriden, no need to use up uidspace
				file->readWarrior(*_warrior[i]);
			}
			delete file;
		} else { // first generation, so create some random warriors
			if(kingdom()->in_experiment()) {
				for(i=0; i<_num_warriors; ++i) {
					_warrior[i] = new CWarrior(this);
					if(!silent) { // debug info
						cout << "\t\tgenerating " << _name << ':' << _warrior[i]->uid() << " .. ";
					}
					tries = 0;
					do {
						tries++;
						_warrior[i]->clear();
						_warrior[i]->set_info("generated from random");
						for(int j=0; j<num_chromosomes(); j++) {
							chromosome(j)->reproduction()->generateCompletelyRandom(*(_warrior[i]->chromosome(j)));
						}
					} while(!CReproduction::fit_enough(*_warrior[i]));
					if(!silent) { // debug info
						cout << "took " << tries << " tries";
						if(verbose) {
							cout << ':' << endl;
							_warrior[i]->rc(cout);
						}
						cout << endl;
					}
				}
			} else { // lazy load?
				for(i=0; i<_num_warriors; ++i) {
					_warrior[i] = 0; // null
				}
			}
		}
	}
}

void CSpecies::write_ini(ostream &os) {
	int i;
	os	<< "uid=" << _uid << endl
		<< "evolving=" << _evolving << endl;
	// dump misc stuff
	_fitness->write_override_ini(os,_genus->fitness());
	_length->write_override_ini(os,_genus->length());
	_freq->write_override_ini(os,_genus->freq());
	_operands->write_override_ini(os,_genus->operands());
	_reproduction->write_override_ini(os,_genus->reproduction());
	// exec trail?
	if(_exec_trail) {
		os << "exec_trail=" << _exec_trail->filename() << endl;
	}
	// fitness weighting?
	if(_fitness_weighting != DEFAULT_FITNESS_WEIGHTING) { // non-default?
		os << "fitness_weighting=" << _fitness_weighting << endl;
	}
	// chromosomes
	if(!_auto_chromosome) {
		os << "num_chromosomes=" << _num_chromosomes << endl;
		for(i=0; i<_num_chromosomes; ++i)
			os << "chromosome_" << i << '=' << _chromosome[i]->name() << endl;
		if(0 != _start_chromosome)
			os << "start_chromosome=" << _start_chromosome << endl;
	}
	// list warriors
	os << "num_warriors=" << size() << endl;
	if(is_bench()) {
		for(i=0; i<_num_warriors; ++i) {
			os << "warrior_" << i << '=' << _warrior[i]->filename() << endl; 
		}
	} else { // is evolver
		save();
	}
	// print each chromosome
	if(!_auto_chromosome) {
		for(i=0; i<_num_chromosomes; ++i)
			_chromosome[i]->write_ini(os);
	}
}

void CSpecies::save() {
	int i;
	CSpeciesFile *file;
	if(is_bench()) // no point?
		return;
	file = new CSpeciesFile(*this,kingdom()->generation());
	for(i=0; i<_num_warriors; ++i) {
		file->writeWarrior(*_warrior[i]);
	}
	delete file;
}

void CSpecies::ready() { /* this will blank all fitness metrics preceeding the fighting of a generation */
	int i;
	for(i=0; i<_num_warriors; ++i) {
		_warrior[i]->ready();
	}
	// resume support
	if(0 != _resume)
		delete _resume;
	if(is_evolving()) {
		_resume = new CResumeFile(this);
		_resuming = _resume->openForRead();
		if(!_resuming)
			_resume->openForWrite();
	} else
		_resuming = false;
}

int CSpecies::fight(CSpecies &enemy) { /* this will actually do all the fights for a generation; returns number of rounds fought */
	CWarrior::FIGHT_DATA data;
	int i, j, count=0, resumed = 0;
	for(i=0; i<_num_warriors; ++i) {
		for(j=0; j<enemy._num_warriors; ++j) {
			// try to resume..
			if(_resuming) {
				if(_resume->read(data)) {
					// sanity check
					if(data.src != _warrior[i]->uid())
						PANIC(BAD_RESUMEFILE,"src uid failed",NULL);
					if(data.dest != enemy._warrior[j]->uid())
						PANIC(BAD_RESUMEFILE,"dest uid failed",NULL);
					// do the actual scoring
					count += _warrior[i]->fight(*enemy._warrior[j],data,false);
					// done, so next enemy
					resumed++;
					continue;
				} else {
					if(!silent && (0 < resumed)) {
						cout << "(loaded " << resumed << " fights from " << name() << "\'s resume log)" << endl;
					}
					// failed, so no more resumes
					_resuming = false;
					_resume->swapFromReadToWrite();
					// drop through
				}
			}
			// actually fight
			PROGRESS_TWIDDLE
			count += _warrior[i]->fight(*enemy._warrior[j],data);
			if(is_evolving()) {
				_resume->write(data);
			}
		}
	}
	if(0 != _resume) {
		_resume->flush();
	}
	if(!silent && _resuming && (0 < resumed)) {
		cout << "(loaded all " << resumed << " fights from " << name() << "\'s resume log)" << endl;
	}
	return count;
}

CWarrior *CSpecies::selection() { /* this will do selection on all the warriors in this species; returns pointer to the best too */
	// there is bound to be a better way than this
	int i, j, tries,
		mother, father, best,
		benchmark_n, benchmark_score, benchmark_sum = 0;
	bool ok;
	fstream red;
	string name, note;
	if(!_evolving) // no point with benchmark warriors
		return 0;
	_resume->close();
	// ready survives table
	for(i=0; i<_num_warriors; ++i) { // clear survives flags
		_survives[i] = false;
	}
	// prepare for benchmarking?
	if(0 != kingdom()->benchmark()) {
		sort();
	}
	// best always survives
	best = scores(_min,_sum);
	_next[0] = _warrior[best];
	_survives[best] = true;
	// benchmark the better ones?
	if(0 != kingdom()->benchmark()) {
		if(verbose) {
			cout << "\tbenchmarking " << _name << endl;
		}
		// how many warriors to test?
		benchmark_n = (int)(_num_warriors * kingdom()->benchmark_n());
		if(0 == benchmark_n) benchmark_n = 1; // always at least the first(best) warrior
		if(benchmark_n > _num_warriors) PANIC(MISC,"bad benchmark_n",NULL);
		// and do it
		for(i=0; i<benchmark_n; ++i) {
			if(!silent) {
				cout << "\t\tbenchmarking " << _name << ':' << _warrior[i]->uid() << " .. " << flush;
			}
			// todo: make this use a variable number of rounds in the benchmark
			benchmark_score = 0;
			for(j=0; j<kingdom()->benchmark()->_num_warriors; ++j) {
				benchmark_score += _warrior[i]->fight(*kingdom()->benchmark()->_warrior[j],100,3,1,0);
			}
			benchmark_score = (int)((double)benchmark_score / (double)kingdom()->benchmark()->num_warriors()); // = score / num_warriors_in_bench / rounds_per_warrior * 100
			if(!silent) {
				cout << "scored " << benchmark_score << endl;
			}
			_warrior[i]->incorporate_benchmark(benchmark_score);
			benchmark_sum += benchmark_score;
			name = _name;
			name += "--";
			string_append(name,_warrior[i]->uid());
			name += ".red";
			note = ";benchmark ";
			string_append(note,(int)round(_warrior[i]->benchmark()));
			if(kingdom()->benchmark_candidate() <= benchmark_score) {
				note += " CANDIDATE!";
				if(!silent) {
					cout << "Warrior " << name << " is a CANDIDATE with a benchmark score of " << benchmark_score << endl;
				}
			}
			red.clear();
			red.open(name.c_str(),ios_base::out|ios_base::trunc);
			if(red) {
				_warrior[i]->red(red,note.c_str());
			} else {
				cerr << "WARNING: could not write to " << name << endl;
			}
			red.close();
		}
	} else { // just save the best of this generation
		name = _name;
		name += "-";
		string_append(name,kingdom()->generation());
		name += ".red";
		red.clear();
		red.open(name.c_str(),ios_base::out|ios_base::trunc);
		if(red) {
			_warrior[best]->red(red);
		} else {
			cerr << "WARNING: could not write to " << name << endl;
		}
		red.close();
	}
	// and adjust the value to return (since best is always moved to _next[0])
	best = 0;
	// and select the rest
	for(i=1; i<_num_warriors; ++i) { // for each new warrior; note base 1
		ok = false;
		do {
			if(reproduction()->survive()) { // does a warrior from the old generation survive?
				j = rnd(); // get a random one
				if(_survives[j]) { // already surviving?
					continue;
				}
				_survives[j] = true;
				_next[i] = _warrior[j];
				if(verbose) {
					cout << _next[i]->uid() << " survives" << endl;
				}
			} else { // breed a warrior
				mother = rnd(); // get a random one
				do {
					father = rnd(); // get a random one
				} while(father == mother); // until different
				_next[i] = new CWarrior(this,_warrior[mother]->uid(),_warrior[father]->uid());
				if(verbose) {
					cout << _next[i]->uid() << " born from " << _warrior[mother]->uid() << " and " << _warrior[father]->uid() << endl;
				}
				tries = 0;
				do {
					tries++;
					for(int j=0; j<num_chromosomes(); j++) {
						chromosome(j)->reproduction()->breed(*(_next[i]->chromosome(j)),*(_warrior[mother]->chromosome(j)),*(_warrior[father]->chromosome(j)));
					}
				} while(!CReproduction::fit_enough(*_next[i]));
				_next[i]->code_check();
				if(verbose) { // debug output
					cout << "generating " << _warrior[i]->uid() << " took " << tries << " tries:" << endl;
					_warrior[i]->rc(cout);
					cout << endl;
				}
			}
			ok = true; // if here, we have generated something
		} while(!ok);
	}
	// transfer next generation to this one
	for(i=0; i<_num_warriors; ++i) {
		if(!_survives[i]) { // clean up?
			delete _warrior[i];
		}
		_warrior[i] = _next[i];
	}
	// done
	return _warrior[best];
}

/*	This returns a random warrior based upon fitness */
int CSpecies::rnd() const {
	int i;
	const CFitness::NUM x = CRand::frand(_sum);
	CFitness::NUM y; fitness()->zero(y);
	do {
		for(i=0; i<_num_warriors; ++i) {
			y += (_warrior[i]->score() - _min); // based upon _min
			if(y > x) { // found it?
				return i;
			}
		}
		cerr << "WARNING: roulette off end of CSpecies for fitness" << endl;
	} while(true);
}

/*	This returns the minimum and sum (adjusted relative to min) scores for this species;
	these are the numbers needed to do roulette selection of the warriors based upon fitness.
	The min is actually 1 less than the true min, so even the worst warriors have a small chance
	of being selected; returns index of the best too */
int CSpecies::scores(CFitness::NUM &min,CFitness::NUM &sum) const {
	int i, ret = 0;
	CFitness::NUM max = _warrior[ret]->score();
	min = max;
	fitness()->zero(sum);
	// find min and max
	for(i=0; i<_num_warriors; ++i) {
		if(_warrior[i]->score() > max) {
			ret = i;
			max = _warrior[ret]->score();
		} else if(_warrior[i]->score() < min) {
			min = _warrior[i]->score();
		}
	}
	// work out sum, adjusted for min as a base
	min--;
	for(i=0; i<_num_warriors; ++i) {
		sum += _warrior[i]->score();
		sum -= _min;
	}
	if(sum <= 0.0F)
		PANIC(MISC,"error in CSpecies::scores(..)",NULL);
	return ret;
}

/* this sorts the warriors into score order, best in 0 */
void CSpecies::sort() {
	// TODO: replace this bubblesort?
	bool changed;
	int i, j, len = _num_warriors;
	CWarrior *swap;
	do {
		changed = false;
		len--;
		j=0;
		for(i=0; i<len; i++) {
			j++;
			if(_warrior[i]->score() < _warrior[j]->score()) {
				// swap the pointers over
				swap = _warrior[i];
				_warrior[i] = _warrior[j];
				_warrior[j] = swap;
				// and remember we've done something
				changed = true;
			}
		}
	} while(changed);
}

CChromosomeType *CSpecies::chromosome(const unsigned int i) const {
	if(safety_checks)
		if(_num_chromosomes < i)
			PANIC(MISC,"chromosome index beyond range",NULL);
	return _chromosome[i];
}

CKingdom *CSpecies::kingdom() const { return _genus->kingdom(); }

CWarrior *CSpecies::warrior(const CWarrior::TUid uid) const { /* fetch a particular warrior from this species; NULL if not present */
	int i;
	CWarrior *ret = 0;
	for(i=0; i<_num_warriors; ++i) {
		if(_warrior[i]->uid() == uid) { // match?
			return _warrior[i];
		}
	}
	return 0; // not found
}

