
#include "visitool.hpp"

/* "Species" - a CoreWars evolver.  Copyright (C) 2003 'Varfar' 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 1, or (at your option) any later
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <math.h>

#include <iostream>
using namespace std;

namespace { // anon
	
	// colours; could be wxColour or strings
	const char 
		*BG_COLOUR = "WHEAT",
		*KINGDOM_LABEL_COLOUR = "WHITE";
		
	wxBrush
		GENUS_BRUSH,
		EVOLVING_SPECIES_BRUSH,
		BENCHMARK_SPECIES_BRUSH;
		
	void init_brushes() {
		// have to do this after the wx system has initialised, else you'll panic
		GENUS_BRUSH = wxBrush("TAN",wxSOLID);
		EVOLVING_SPECIES_BRUSH = wxBrush("WHITE",wxSOLID);
		BENCHMARK_SPECIES_BRUSH = wxBrush("YELLOW",wxSOLID);
	}
	
} // anon namespace

// useful defines
#define OPEN_PANIC_TRAP try {
#define CLOSE_PANIC_TRAP_FATAL }\
	catch(CSpeciesError *err) {\
		wxMessageBox(err->desc().c_str(),"An error occured",wxOK|wxICON_ERROR,NULL);\
		delete err;\
		exit(1);\
	}
#define CLOSE_PANIC_TRAP_WARN }\
	catch(CSpeciesError *err) {\
		wxMessageBox(err->desc().c_str(),"An error occured",wxOK|wxICON_WARNING,NULL);\
		delete err;\
	}
	
/************ SimulationThread implementation *************************/

SimulationThread::SimulationThread(SpeciesFrame *owner): wxThread(wxTHREAD_DETACHED) {
	_owner = owner;
	_stopping = false;
	_stopped = true;
	// create kingdom
	OPEN_PANIC_TRAP
	_kingdom = new CKingdom();
	_kingdom->load(false); // lazy
	CLOSE_PANIC_TRAP_FATAL
}

SimulationThread::~SimulationThread() {
	delete _kingdom;
}

wxThread::ExitCode SimulationThread::Entry() {
	_stopped = false;
	OPEN_PANIC_TRAP
	/* kingdom.start() */
	while(!_stopping) {
		_kingdom->fight();
	}
	CLOSE_PANIC_TRAP_FATAL
	_stopped = true;
	if(0 != _owner)
		_owner->OnSimulationStopped();
	return NULL;
}

/************ SpeciesFrame implementation *****************************/

BEGIN_EVENT_TABLE(SpeciesFrame, wxFrame)
  EVT_MENU(SpeciesFrame::ID_Quit,SpeciesFrame::OnMnuFileQuit)
  EVT_MENU(SpeciesFrame::ID_About,SpeciesFrame::OnMnuFileAbout)
  EVT_PAINT(SpeciesFrame::OnPaint)
  EVT_SIZE(SpeciesFrame::OnSize)
  EVT_BUTTON(SpeciesFrame::ID_Run_Simulation,SpeciesFrame::OnRunSimulation)
  EVT_BUTTON(SpeciesFrame::ID_Stop_Simulation,SpeciesFrame::OnStopSimulation)
END_EVENT_TABLE()

// constructor
SpeciesFrame::SpeciesFrame(const wxString& title, const wxPoint& pos, const wxSize& size):
	wxFrame((wxFrame *)NULL,-1,title,pos,size) {
	// create simulation thread
	_simthread = new SimulationThread(this);
	if(wxTHREAD_NO_ERROR != _simthread->Create(10000)) {
		wxMessageBox("Could not create Simulation Thread","An error occured",wxOK|wxICON_ERROR,NULL);
		exit(1);
	}
	// make menu
	wxMenu *menuFile = new wxMenu;
	menuFile->Append(ID_Quit,"E&xit");
	wxMenuBar *menuBar = new wxMenuBar;
	menuBar->Append(menuFile,"&File");
	menuFile = new wxMenu;
	menuFile->Append(ID_About,"&About...");
	menuBar->Append(menuFile,"&Help");
	SetMenuBar(menuBar);
	// make kingdom label
	cout << "ok" << endl << flush;
	wxString lblKingdom = "test";
	OPEN_PANIC_TRAP
	lblKingdom.Printf("Kingdom\n\n%i Genera\n%i Evolving Species\n%i Evolved Warriors",kingdom()->num_genera(),kingdom()->num_evolving_species(),kingdom()->num_evolved_warriors());
	CLOSE_PANIC_TRAP_WARN
	_kingdom_label = new wxStaticText(this,-1,lblKingdom,wxPoint(5,5));
	_kingdom_label->SetBackgroundColour(KINGDOM_LABEL_COLOUR);
	// make run/stop simulation buttons
	_run_stop_panel = new wxPanel(this,-1,wxDefaultPosition,wxSize(100,20));
	wxSizer *run_stop_sizer = new wxBoxSizer(wxHORIZONTAL);
	_run_simulation = new wxButton(_run_stop_panel,ID_Run_Simulation,"Run");
	run_stop_sizer->Add(_run_simulation,1,wxEXPAND|wxADJUST_MINSIZE|wxALL,3);
	_stop_simulation = new wxButton(_run_stop_panel,ID_Stop_Simulation,"Stop");
	_stop_simulation->Disable();
	run_stop_sizer->Add(_stop_simulation,1,wxEXPAND|wxADJUST_MINSIZE|wxTOP|wxRIGHT|wxBOTTOM,3);
	_run_stop_panel->SetSizer(run_stop_sizer);
	run_stop_sizer->SetSizeHints(_run_stop_panel);
	run_stop_sizer->Layout();
	_run_stop_panel->SetBackgroundColour(BG_COLOUR);
	_run_stop_panel->SetAutoLayout(TRUE);
	// misc stuff
	SetBackgroundColour(BG_COLOUR);
}
	
//*** event listeners

void SpeciesFrame::OnRunSimulation() {
	_run_simulation->Disable();
	_stop_simulation->Enable();
	_simthread->Run();
}

void SpeciesFrame::OnStopSimulation() {
	_run_simulation->Disable();
	_stop_simulation->Disable();
	_simthread->Stop();
}

void SpeciesFrame::OnSimulationStopped() {
	_run_simulation->Enable();
	_stop_simulation->Disable();
}

void SpeciesFrame::OnMnuFileAbout(wxCommandEvent& WXUNUSED(event)) {
	wxMessageBox("This tool allows easy examination of Species evolver status\n"
		"For details about Species, see http://redcoder.sourceforge.net\n\n"
		"(c) Will 'Varfar', 2003; This is GPL software",
		"About Species Visitool",wxOK|wxICON_INFORMATION,this);
}

void SpeciesFrame::OnMnuFileQuit(wxCommandEvent& WXUNUSED(event)) {
	Close(TRUE);
}

void SpeciesFrame::OnPaint(wxPaintEvent& WXUNUSED(event)) {
	wxPaintDC dc(this);
	dc.BeginDrawing();
	draw(dc);
	dc.EndDrawing();
}

void SpeciesFrame::OnSize(wxSizeEvent& WXUNUSED(event)) {
	if(!IsShown()) return;
	calcLayout();
	// put the run/stop buttons in the right-hand-corner
	wxSize inner = GetClientSize(), run_stop = _run_stop_panel->GetSize();
	_run_stop_panel->SetSize(inner.x-run_stop.x,inner.y-run_stop.y,run_stop.x,run_stop.y);
}
	
void SpeciesFrame::draw(wxDC& dc) {
	dc.SetBrush(GENUS_BRUSH);
	for(int i=0; i<kingdom()->num_genera(); i++) {
		const CGenus &genus = kingdom()->genus(i);
		CircleTag *hook = (CircleTag*)(genus.hook());
		if(0 != hook) {
			hook->draw(dc);
			for(int j=0; j<genus.num_species(); j++) {
				hook = (CircleTag*)(genus.species(j).hook());
				if(0 != hook) {
					dc.SetBrush(genus.species(j).is_evolving() ? EVOLVING_SPECIES_BRUSH: BENCHMARK_SPECIES_BRUSH);
					hook->draw(dc);
				}
			}
			dc.SetBrush(GENUS_BRUSH);
		}
	}
}

SpeciesFrame::CircleTag::CircleTag(const wxString &label,const int nitems,const int index,const wxRect &bounds) {
	const Precision PI = 3.1415926535898,
		RADS_IN_CIRCLE = 2 * PI;
	// work out centre of containing bounds
	int half_width = bounds.GetWidth() / 2,
		half_height = bounds.GetHeight() / 2;
	Precision dist_from_centre = (half_width < half_height) ? half_width: half_height; // bounding box might not be square
	// and work out where we fit in
	Precision c = (RADS_IN_CIRCLE / (Precision)nitems),
		theta = c * (Precision)index;
	// special case?
	if(1 == nitems) {
		c /= 2; // pretend there are two items
	}
	// work out radius; THANKYOU JOONAS!!
	_radius = (wxCoord)round(dist_from_centre - (dist_from_centre / (1.0 + sin(c / 2.0))));
	// and work out where to put the circle
	dist_from_centre -= _radius;
	_radius = (wxCoord)round((Precision)_radius * 0.9); // shrink it slightly
	// actually calc position of circle
	_x = (bounds.GetX() + half_width) + (int)round(cos(theta) * dist_from_centre);
	_y = (bounds.GetY() + half_height) + (int)round(sin(theta) * dist_from_centre);
	// and work out my bounds
	_bounds.SetX(_x - _radius);
	_bounds.SetY(_y - _radius);
	_bounds.SetWidth(_radius * 2);
	_bounds.SetHeight(_radius * 2);
	// misc
	_label = label;
}

void SpeciesFrame::CircleTag::draw(wxDC &dc) {
	dc.DrawCircle(_x,_y,_radius);
	// work out label
	wxCoord tw, th;
	dc.GetTextExtent(_label,&tw,&th);
	dc.DrawText(_label,_x - (tw / 2),_y - (th / 2));
}

void SpeciesFrame::calcLayout() {
	OPEN_PANIC_TRAP
	const int n = kingdom()->num_genera();
	wxRect bounds(GetClientAreaOrigin(),GetClientSize());
	for(int i=0; i<n; i++) {
		CGenus &genus = kingdom()->genus(i);
		CircleTag *hook = new CircleTag(genus.name().c_str(),n,i,bounds);
		genus.hook(hook);
		const int m = genus.num_species();
		for(int j=0; j<m; j++) {
			genus.species(j).hook(new CircleTag(genus.species(j).name().c_str(),m,j,hook->bounds()));
		}
	}
	CLOSE_PANIC_TRAP_FATAL
}

/************ VisiToolApp implementation *******************************/

IMPLEMENT_APP(VisiToolApp)

bool VisiToolApp::OnInit() {
	init_brushes();
	// create frame
	SpeciesFrame *frame = new SpeciesFrame( "Species - A Corewars Evolver", wxPoint(50,50), wxSize(450,340) );
	frame->CentreOnScreen(wxBOTH); // for when not maximised
	frame->Maximize(TRUE);
	// and show it
	frame->Show(TRUE);
	SetTopWindow(frame);
	// done
	return TRUE;
}

